/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_DISTRIBUTION_PARAMETER_PARAMETER_PACK_HPP
#define GRAPH_MODEL_DISTRIBUTION_PARAMETER_PARAMETER_PACK_HPP 1

#include <type_traits>
#include <boost/hana/append.hpp>
#include <boost/hana/at_key.hpp>
#include <boost/hana/contains.hpp>
#include <boost/hana/drop_while.hpp>
#include <boost/hana/filter.hpp>
#include <boost/hana/insert.hpp>
#include <boost/hana/integral_constant.hpp>
#include <boost/hana/length.hpp>
#include <boost/hana/map.hpp>
#include <boost/hana/minus.hpp>
#include <boost/hana/not_equal.hpp>
#include <boost/hana/tuple.hpp>
#include <always_false.hpp>
#include <graph_model/distribution/parameter/named_parameter.hpp>
#include <graph_model/distribution/parameter/parameter.hpp>
#include <graph_model/vertex/expression_function.hpp>

namespace graph_model
{

  namespace distribution
  {

    namespace parameter
    {

      /// \brief A distribution parameter pack
      /// \tparam Tags A tuple of valid tags
      /// \tparam Defaults A map of tag/value pairs for defaults (may be empty)
      /// \tparam Tagged A map of tag/value pairs for explicitly tagged parameters (may be empty)
      /// \tparam Untagged A tuple of untagged parameter values (may be empty)
      ///
      /// A parameter pack enables looking up parameter values by \c
      /// tag, which must be within the tuple of valid \tparam Tags.
      /// The lookup algorithm is the following:
      ///
      /// - if the \c tag is in \c keys(Tagged), then the value is \c
      ///   Tagged[tag]
      ///
      /// - otherwise
      ///
      ///   - let \c T = ordered set of tags in \tparam Tags that are
      ///     not in \c keys(Tagged), i.e., \c
      ///     filters(keys(Tagged),tag_not_in_Tagged())
      ///
      ///   - let \c I = \c index_of(T,tag)
      ///
      ///   - if \c I < \c size(Untagged), then the value is \c
      ///     Untagged[I]
      ///
      ///   - otherwise if \c tag is in \tparam Defaults, the value is \c Defaults[tag]
      ///
      ///   - otherwise, compile error
      ///   
      template < typename Tags = boost::hana::tuple<>
		 , typename Defaults = boost::hana::tuple<>
		 , typename Tagged = boost::hana::tuple<>
		 , typename Untagged = boost::hana::tuple<>
		 >
      class parameter_pack
      {
	using tags_type = Tags;
	using defaults_type = Defaults;
	using tagged_type = Tagged;
	using untagged_type = Untagged;

	using key_is_in_tagged = std::true_type;
	using key_is_not_in_tagged = std::false_type;

	using index_is_in_untagged = std::true_type;
	using index_is_not_in_untagged = std::false_type;

	using key_is_in_defaults = std::true_type;
	using key_is_not_in_defaults = std::false_type;

	template < typename Key >
	static auto enable_if_tags_contains (Key)
	{
	  return std::enable_if_t
	    <decltype(boost::hana::contains(std::declval<Tags>(),boost::hana::type_c<Key>))::value>();
	}

	template < typename Container, typename Key >
	static auto container_contains_key (Container container, Key)
	{
	  return std::integral_constant
	    <bool,decltype(boost::hana::contains(container,boost::hana::type_c<Key>))::value>();
	}

	template < typename Container >
	static auto container_is_empty (Container container)
	{ return std::integral_constant<bool,decltype(boost::hana::length(container))::value == 0>(); }

	template < typename Container, typename Index >
	static auto index_is_valid (Container container, Index index)
	{ return std::integral_constant<bool,index.value < decltype(boost::hana::length(container))::value>(); }

	auto tag_not_in_Tagged () const
	{ return [tagged=tagged_](auto tag){ return !boost::hana::contains(tagged,tag); }; }

	template < typename Iterable, typename T >
	static constexpr auto index_of (Iterable const& iterable, T const& element)
	{
	  auto length = decltype(boost::hana::length(iterable)){};
	  auto dropped = decltype
	    (boost::hana::length
	     (boost::hana::drop_while(iterable, boost::hana::not_equal.to(element))
	      )){};
	  return length - dropped;
	}

	template < typename Key >
	auto index_of_key_in_remaining_tags (Key) const
	{
	  using boost::hana::filter;
	  auto T = filter(tags_,tag_not_in_Tagged());
	  return index_of(T,boost::hana::type_c<Key>);
	}

	template < typename Map, typename Key >
	static auto map_value (Map& map, Key key)
	{
	  using boost::hana::type_c;
	  return map[type_c<Key>];
	}

      private:
	struct append_map
	{
	  template < typename Key, typename Value >
	  static auto apply (boost::hana::tuple<>, Value value)
	  {
	    using boost::hana::make_map;
	    using boost::hana::make_pair;
	    using boost::hana::type_c;
	    return make_map(make_pair(type_c<Key>,value));
	  }

	  template < typename Key, typename Map, typename Value >
	  static auto apply (Map map, Value value)
	  {
	    using boost::hana::insert;
	    using boost::hana::make_map;
	    using boost::hana::make_pair;
	    using boost::hana::type_c;

	    using key_already_exists = decltype(container_contains_key(map,Key()));
	    static_assert(!key_already_exists::value,"Key already exists in a parameter pack");
	    return boost::hana::insert(map,make_pair(boost::hana::type_c<Key>,value));
	  }
	};

	struct append_tuple
	{
	  template < typename Container, typename Value >
	  static auto apply (Container container, Value value)
	  {
	    using boost::hana::append;
	    return append(container,value);
	  }
	};

      public:
	parameter_pack () = default;
	parameter_pack (Tags tags, Defaults defaults, Tagged tagged, Untagged untagged)
	  : tags_(tags)
	  , defaults_(defaults)
	  , tagged_(tagged)
	  , untagged_(untagged)
	{}

	auto add () const { return *this; }

	template < typename Key, typename ... Parameters >
	auto add (parameter<Key,void> parameter, Parameters&& ... parameters) const
	{
	  auto tags = append_tuple::apply(tags_,boost::hana::type_c<Key>);
	  using parameter_pack_type = parameter_pack<decltype(tags),Defaults,Tagged,Untagged>;
	  return parameter_pack_type(tags,defaults_,tagged_,untagged_)
	    .add(std::forward<Parameters>(parameters)...);
	}

	template < typename Key, typename Expr, typename ... Parameters >
	auto add (parameter<Key,Expr> parameter, Parameters&& ... parameters) const
	{
	  auto tags = append_tuple::apply(tags_,boost::hana::type_c<Key>);
	  auto defaults = append_map::template apply<Key>(defaults_,parameter());
	  using parameter_pack_type = parameter_pack<decltype(tags),decltype(defaults),Tagged,Untagged>;
	  return parameter_pack_type(tags,defaults,tagged_,untagged_)
	    .add(std::forward<Parameters>(parameters)...);
	}

	template < typename Tag, typename Value, typename ... Parameters >
	auto add (named_parameter<Tag,Value> value, Parameters&& ... parameters)
	{ return add<Tag>(value()).add(std::forward<Parameters>(parameters)...); }

	template < typename Key
		   , typename Value
		   , typename = decltype(enable_if_tags_contains(Key()))
		   >
	auto add (Value value) const
	{
	  auto tagged = append_map::template apply<Key>(tagged_,value);
	  using parameter_pack_type = parameter_pack<Tags,Defaults,decltype(tagged),Untagged>;
	  return parameter_pack_type(tags_,defaults_,tagged,untagged_);
	}

	template < typename Value, typename ... Parameters >
	auto add (Value&& value, Parameters&& ... parameters) const
	{
	  using boost::hana::length;
	  using vertex::make_expression_function;
	  auto function = make_expression_function(std::forward<Value>(value));
	  auto untagged = append_tuple::apply(untagged_,function);
	  using parameter_pack_type = parameter_pack<Tags,Defaults,Tagged,decltype(untagged)>;
	  return parameter_pack_type(tags_,defaults_,tagged_,untagged)
	    .add(std::forward<Parameters>(parameters)...);
	}

      public:
	template < typename Key
		   , typename ... Args
		   , typename = decltype(enable_if_tags_contains(Key()))
		   >
	auto value (Args&& ... args) const { return get_value<Key>()(std::forward<Args>(args)...); }

      private:
	template < typename Key >
	auto get_value () const
	{
	  auto key_is_in_tagged = container_contains_key(tagged_,Key());
	  return get_value<Key>(key_is_in_tagged);
	}

	template < typename Key >
	auto get_value (key_is_in_tagged) const
	{
	  using boost::hana::type_c;
	  return tagged_[type_c<Key>];
	}

	template < typename Key >
	auto get_value (key_is_not_in_tagged) const
	{
	  auto index = index_of_key_in_remaining_tags(Key());
	  auto index_is_in_untagged = index_is_valid(untagged_,index);
	  return get_value<Key>(key_is_not_in_tagged(),index_is_in_untagged,index);
	}

	template < typename Key, typename Index >
	auto get_value (key_is_not_in_tagged, index_is_in_untagged, Index index) const
	{
	  return boost::hana::at_c<index.value>(untagged_);
	}

	template < typename Key, typename Index >
	auto get_value (key_is_not_in_tagged, index_is_not_in_untagged, Index index) const
	{
	  using boost::hana::type_c;
	  auto key_is_in_defaults = container_contains_key(defaults_,Key());
	  return get_value<Key>(key_is_not_in_tagged(),index_is_not_in_untagged(),key_is_in_defaults,index);
	}

	template < typename Key, typename Index >
	auto get_value (key_is_not_in_tagged, index_is_not_in_untagged, key_is_in_defaults, Index) const
	{
	  using boost::hana::type_c;
	  return defaults_[type_c<Key>];
	}

	template < typename Key, typename Index >
	auto get_value (key_is_not_in_tagged, index_is_not_in_untagged, key_is_not_in_defaults, Index) const
	{
	  static_assert(always_false<Index>::value,
			"No value for Key; "
			"it is not in Tagged, "
			"does not correspond to a valid index in Untagged, "
			"and is not in Defaults");
	}

      private:
	tags_type tags_;
	defaults_type defaults_;
	tagged_type tagged_;
	untagged_type untagged_;
      };

      auto make_parameter_pack () { return parameter_pack<>(); }

      template < typename ... Parameters >
      auto make_parameter_pack (Parameters&& ... parameters)
      { return make_parameter_pack().add(std::forward<Parameters>(parameters)...); }

    } // namespace parameter

  } // namespace distribution

} // namespace graph_model

#endif
