/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_UTILITY_FUNCTION_COLLECTION_HPP
#define GRAPH_MODEL_UTILITY_FUNCTION_COLLECTION_HPP 1

#include <boost/poly_collection/function_collection.hpp>
#include <graph_model/type_traits/has_equal_to.hpp>
#include <graph_model/type_traits/has_less.hpp>

namespace graph_model
{

  namespace utility
  {

    /// \brief Collection of function objects
    /// \tparam Signature The function signature
    ///
    /// This class maintains a collection of callable objects.  It is
    /// built upon the Boost.PolyCollection library to add support for
    /// stateless lambdas, which are not correctly handled by \c
    /// boost::has_less.  Presumably, this will not be needed once
    /// Boost.PolyCollection handles stateless lambdas properly.
    ///
    template < typename Signature >
    class function_collection;

    template < typename Result, typename ... Args >
    class function_collection< Result(Args...) >
      : public boost::poly_collection::function_collection<Result(Args...)>
    {
      using super = boost::poly_collection::function_collection<Result(Args...)>;
      using super::super;

      template < typename F >
      class callable_wrapper
      {
	using self = callable_wrapper;

      public:
	callable_wrapper (F const& f) : f_(f) {}
	callable_wrapper (F&& f) : f_(std::move(f)) {}

	Result operator () (Args ... args) const { return invoke(args...); }

	friend bool operator == (self const& left, self const& right) { return left.equal_to(right); }
	friend bool operator < (self const& left, self const& right) { return left.less(right); }

      private:
	Result invoke (Args ... args) const
	{
	  using result_is_void = typename std::is_void<Result>::type;
	  return invoke(result_is_void(),args...);
	}

	Result invoke (std::false_type, Args ... args) const { return f_(args...); }
	Result invoke (std::true_type, Args ... args) const { f_(args...); }

      private:
	bool equal_to (self const& other) const
	{
	  using has_equal_to = type_traits::has_equal_to_t<F>;
	  return equal_to(has_equal_to(),other);
	}

	bool equal_to (std::false_type, self const& other) const { return false; }
	bool equal_to (std::true_type, self const& other) const { return f_ == other.f_; }

      private:
	bool less (self const& other) const
	{
	  using has_less = type_traits::has_less_t<F>;
	  return less(has_less(),other);
	}

	bool less (std::false_type, self const& other) const { return false; }
	bool less (std::true_type, self const& other) const { return f_ < other.f_; }

      private:
	F f_;
      };

    public:
      auto begin () const { return super::begin(); }
      auto end () const { return super::end(); }

      template < typename F >
      auto begin () const
      {
	using callable_wrapper = callable_wrapper<std::decay_t<F>>;
	return super::template begin<callable_wrapper>();
      }

      template < typename F >
      auto end () const
      {
	using callable_wrapper = callable_wrapper<std::decay_t<F>>;
	return super::template end<callable_wrapper>();
      }

    public:
      template < typename F >
      auto insert (F&& f)
      {
	using callable_wrapper = callable_wrapper<std::decay_t<F>>;
	return super::insert(callable_wrapper(f));
      }

    protected:
      template < typename F >
      auto insert_unique (F&& f)
      {
	using callable_wrapper = callable_wrapper<std::decay_t<F>>;
	callable_wrapper wrapper(std::forward<F>(f));
	super::template register_types<callable_wrapper>();
	auto first = super::template begin<callable_wrapper>(),
	  last = super::template end<callable_wrapper>(),
	  i = std::lower_bound(first,last,wrapper);
	if (i != last && !(wrapper < *i))
	  return i;
	return super::insert(i,wrapper);
      }
    };

  } // namespace utility

} // namespace graph_model

#endif
