/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_VERTEX_IDENTIFIER_IMPL_BASE_HPP
#define GRAPH_MODEL_VERTEX_IDENTIFIER_IMPL_BASE_HPP 1

namespace graph_model
{

  namespace vertex
  {

    /// \brief Allow chaining of \c reset_identifier()
    ///
    /// Because the \c identifier_base class is part of a \c
    /// vertex_descriptor, it cannot depend on derived classes and
    /// cannot participate in the CRTP.  Consequently, its \c
    /// reset_identifier() member function cannot return a reference
    /// to the derived class and thus cannot be chained with other
    /// member functions.  To overcome this inconvenience, this class
    /// raises the \c reset_identifier() member function, which is
    /// delegated to the \c Base class, to a level that can
    /// participate in the CRTP and therefore can return a reference
    /// to the derived class.  This allows chaining operations on
    /// derived classes.
    ///
    template < typename Derived, typename Base >
    class identifier_impl_base
      : public Base
    {
      using base = Base;

    public:
      template < typename I, typename ... Is >
      auto& reset_identifier (I identifier, Is ... identifiers)
      {
	base::reset_identifier(identifier,identifiers...);
	return derived();
      }

    private:
      auto& derived () { return static_cast<Derived&>(*this); }
      auto& derived () const { return static_cast<Derived const&>(*this); }
    };

  } // namespace vertex

} // namespace graph_model

#endif
