/*
 * Copyright (c) 2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef GRAPH_MODEL_VERTEX_MARKOV_BLANKET_IMPL_BASE_HPP
#define GRAPH_MODEL_VERTEX_MARKOV_BLANKET_IMPL_BASE_HPP 1

#include <boost/log/sources/severity_logger.hpp>
#include <boost/log/trivial.hpp>
#include <asynchronous/signals/product_combiner.hpp>
#include <graph_model/signals/signal.hpp>
#include <graph_model/signals/tags.hpp>
#include <graph_model/type/signal.hpp>
#include <graph_model/type_traits/is_vertex.hpp>
#include <graph_model/vertex/expression_function.hpp>
#include <logging/global_log.hpp>
#include <logging/channel_log.hpp>
#include <mpl/enable_constructor.hpp>
#include <use_default.hpp>

namespace graph_model
{

  namespace vertex
  {

    /// \brief Manage the terms of a Markov blanket
    /// \tparam Derived The type of derived class for CRTP
    /// \tparam Base The base class
    ///
    /// Terms of a Markov blanket are functions with the signature \c
    /// typename \c base::probability_type().  The terms are stored in
    /// a signal and can be added to the Markov blanket at runtime
    /// with overloads of the \c add_factor() member function.  The \c
    /// markov_blanket() member function calculates the value of the
    /// Markov blanket.
    ///
    template < typename Derived
	       , typename Base
	       >
    class markov_blanket_impl_base
      : public Base
    {
      using self = markov_blanket_impl_base;
      using base = Base;

      using signature = typename base::probability_type();
      using signal_tag = signals::tag::markov_blanket;
      using combiner_type = asynchronous::signals::product_combiner<typename base::probability_type>;
      using markov_blanket_signal_type = ::graph_model::type::signal_t<signature,combiner_type,signal_tag>;
      using log_type = logging::channel_log;

    public:
      auto& reset_invocation_strategy () { signal_.reset_invocation_strategy(); return *this; }

      template < typename InvocationPolicy >
      auto& reset_invocation_strategy (InvocationPolicy&& invocation_strategy)
      { signal_.reset_invocation_strategy(std::forward<InvocationPolicy>(invocation_strategy)); return *this; }

      template < typename Wrapper >
      auto& add_factor (expression_function<Wrapper>&& function)
      {
	using namespace boost::log::trivial;
	GLOBAL_LOG(log_,trace) << "add_factor<Wrapper>(expression_function<Wrapper>&&) [" << this << "]: identifier=" << derived().identifier();
	signal_.connect(std::move(function));
	return derived();
      }

      template < typename Expr >
      auto& add_factor (Expr&& expr)
      {
	using namespace boost::log::trivial;
	GLOBAL_LOG(log_,trace) << "add_factor<Expr>(Expr&&) [" << this << "]: identifier=" << derived().identifier();
	using is_vertex = type_traits::is_vertex_t<Expr>;
	add_factor_vertex(std::forward<Expr>(expr),is_vertex());
	return add_factor(make_expression_function(std::forward<Expr>(expr)));
      }

    private:
      template < typename Expr >
      void add_factor_vertex (Expr&& expr, std::false_type)
      {
	using namespace boost::log::trivial;
	GLOBAL_LOG(log_,trace) << "add_factor_vertex<Expr>(Expr&&,std::false_type) [" << this << "]: identifier=" << derived().identifier();
      }

      template < typename Expr >
      void add_factor_vertex (Expr&& expr, std::true_type)
      {
	using namespace boost::log::trivial;
	GLOBAL_LOG(log_,trace) << "add_factor_vertex<Expr>(Expr&&,std::true_type) [" << this << "]: identifier=" << derived().identifier();
	derived().add_vertex(std::forward<Expr>(expr));
      }

    private:
      typename base::markov_blanket_size_type markov_blanket_size_ () const override { return signal_.num_slots(); }

      typename base::probability_type markov_blanket_ () const override
      {
	using namespace boost::log::trivial;
	GLOBAL_LOG(log_,trace) << "markov_blanket_()const override [" << this << "]: identifier=" << derived().identifier();
	auto result = signal_().result();
	GLOBAL_LOG(log_,trace) << "markov_blanket_()const override [" << this << "]: identifier=" << derived().identifier() << ", result=" << result;
	return result;
      };	

    private:
      auto& derived () { return static_cast<Derived&>(*this); }
      auto& derived () const { return static_cast<Derived const&>(*this); }

    private:
      markov_blanket_signal_type signal_;
      mutable log_type log_ = "graph_model::MCMC::markov_blanket_impl_base";
    };

  } // namespace vertex

} // namespace graph_model

#endif
