/*
 * Copyright (c) 2016,2017 Brook Milligan.
 *
 * This file is part of the GraphModel Library.
 *
 *     The GraphModel Library is free software: you can redistribute
 *     it and/or modify it under the terms of the GNU Affero General
 *     Public License as published by the Free Software Foundation,
 *     either version 3 of the License, or (at your option) any later
 *     version.
 *
 *     The GraphModel Library is distributed in the hope that it will
 *     be useful, but WITHOUT ANY WARRANTY; without even the implied
 *     warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 *     PURPOSE.  See the GNU Affero General Public License for more
 *     details.
 *
 *     You should have received a copy of the GNU Affero General
 *     Public License along with the GraphModel Library.  If not, see
 *     <http://www.gnu.org/licenses/>.
 */

/// \file

#ifndef MCMC_MCMC_TALLY_HPP
#define MCMC_MCMC_TALLY_HPP 1

#include <iosfwd>
#include <mcmc/mcmc_tally_fwd.hpp>
#include <observer_ptr.hpp>

namespace mcmc
{
  
  template < typename Count, typename Policy >
  class mcmc_tally
  {
    class acceptor
    {
      using pointer = observer_ptr<mcmc_tally>;

    public:
      acceptor (mcmc_tally& p) : p_(make_observer(p)) {}
      template < typename ... Args >
      void operator () (Args&&...) const { p_->accept_one(); }
    private:
      pointer p_;
    };

    class rejector
    {
      using pointer = observer_ptr<mcmc_tally>;

    public:
      rejector (mcmc_tally& p) : p_(make_observer(p)) {}
      template < typename ... Args >
      void operator () (Args&&...) const { p_->reject_one(); }
    private:
      pointer p_;
    };

    void accept_one () { ++accepted_; policy_(*this); }
    void reject_one () { ++rejected_; policy_(*this); }

  public:
    mcmc_tally() = default;
    mcmc_tally (Policy policy) : policy_(policy) {};
    
    Count total () const { return accepted_ + rejected_; }
    Count accepted () const { return accepted_; }
    Count rejected () const { return rejected_; }
    Count percent_accepted () const { return total() > 0 ? 100 * accepted() / total() : 0; }
    Count percent_rejected () const { return total() > 0 ? 100 * rejected() / total() : 0; }
  
    acceptor make_acceptor () { return acceptor(*this); }
    rejector make_rejector () { return rejector(*this); }

    void reset () { accepted_ = 0; rejected_ = 0; }
  
  public:

    friend std::ostream& operator << (std::ostream& os, mcmc_tally tally)
    {
      return
	os << tally.accepted() << "/" << tally.total() << " (" << tally.percent_accepted() << "%) accepted, "
	   << tally.rejected() << "/" << tally.total() << " (" << tally.percent_rejected() << "%) rejected";
    }
  
  private:
    Count accepted_ = 0;
    Count rejected_ = 0;
    Policy policy_ ;
  };

  auto make_mcmc_tally () { return mcmc_tally<>(); }

} // namespace mcmc

#endif
